/*
 * Copyright (c) 2012-2013 University of Utah.
 * All Rights Reserved.
 *
 * University of Utah grants permission to use, copy, modify, and distribute this software and
 * its documentation for NON-COMMERCIAL purposes and without fee, provided that
 * this copyright notice appears in all copies.
 *
 * University of Utah provides this software "as is," without representations or warranties of
 * any kind, either expressed or implied, including but not limited to the
 * implied warranties of merchantability, fitness for a particular purpose, and
 * noninfringement.  University of Utah shall not be liable for any damages arising from any
 * use of this software.
 *
 * Author: Yan Zheng (yanzheng@cs.utah.edu)
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include "headers.h"
#include "mtimer.cpp"
//#include "ComputeKernel.h"
#include "org_myorg_HadoopKernel.h"

// The data set containing all the points.
PPointT *dataSetPoints = NULL;
// Number of points in the data set.
int nPoints = 0;
// The dimension of the points.
int pointsDimension = 0;
// float

PPointMatchStructT pmStruct;

// Prints the usage.
void usage(char *programName){
	printf("Usage: %s #pts_in_data_set dimension data_set_file epsilon outfile stdv\n", programName);
}

//read a point from file
inline PPointT readPoint(FILE *fileHandle){
  	PPointT p;
  	float sqrLength = 0;
  	FAILIF(NULL == (p = (PPointT)MALLOC(sizeof(PointT))));
  	FAILIF(NULL == (p->coordinates = (float*)MALLOC(pointsDimension * sizeof(float))));
  	for(int d = 0; d < pointsDimension; d++){
		int rv = fscanf(fileHandle, "%f", &(p->coordinates[d]));
 	}
  	p->index = -1;
  	return p;
}

//read the data from file and at the same time to store the min and max in each dimension
void readDataSetFromFile(char *filename, float* minarr, float* maxarr){
 	 FILE *f = fopen(filename, "rt");
  	FAILIF(f == NULL);
	float* cor;
  	FAILIF(NULL == (dataSetPoints = (PPointT*)MALLOC(nPoints * sizeof(PPointT))));
  	for(int i = 0; i < nPoints; i++){
    	dataSetPoints[i] = readPoint(f);
    	dataSetPoints[i]->index = i;
		cor = dataSetPoints[i]->coordinates;
        if(i == 0){
            for(int j = 0; j<pointsDimension; j++){
                minarr[j] = cor[j];
                maxarr[j] = cor[j];
            }
        }else{
            for(int j = 0; j<pointsDimension; j++){
                if(cor[j] < minarr[j]){
                    minarr[j] = cor[j];
                }
                if(cor[j] > maxarr[j]){
                    maxarr[j] = cor[j];
                }
            }
      	}
  }
}

//free space
void freeDataPoints(PPointT* dataSetPoints, int nPoints){
	PPointT p;
	for(int i = 0; i<nPoints; i++){
		p = dataSetPoints[i];
		if(p){
			free(p->coordinates);
			free(p);
		}
	}
	free(dataSetPoints);
}

//output the sample to a file
void getFloatArrSample( PPointT *dataSet,int nPoints, int dim, char* file){
    FILE *fp = fopen(file,"w");
    for(int i = 0; i<nPoints; i++){
        for(int j = 0; j<dim; j++){
            fprintf(fp,"%.1f ",dataSet[i]->coordinates[j]);
        }
        fprintf(fp,"\n");
    }
    fclose(fp);
}


float** getFloatArrSample(PPointT *dataSet,int nPoints, int dim){
    float** sample;
    FAILIF(NULL == (sample = (float**)MALLOC(sizeof(float*)*nPoints)));
    for(int i = 0; i<nPoints; i++){
        FAILIF(NULL == (sample[i] = (float*)MALLOC(sizeof(float)*dim)));
        for(int j = 0; j<dim; j++){
            sample[i][j] = dataSet[i]->coordinates[j];
        }
    }
    return sample;
}

inline PPointT readPointFloat(float *singledata){
    PPointT p;
    FAILIF(NULL == (p = (PPointT)MALLOC(sizeof(PointT))));
    p->coordinates = singledata;
    p->index = -1;
    return p;
}

void readDataSetFromArray(float** data, int nRSample, float* minarr, float* maxarr){
    float* cor;
    FAILIF(NULL == (dataSetPoints = (PPointT*)MALLOC(nRSample * sizeof(PPointT))));
    for(int i = 0; i < nRSample; i++){
        //printf("i = %d\n",i); 
       dataSetPoints[i] = readPointFloat(data[i]);
        dataSetPoints[i]->index = i;
        dataSetPoints[i]->tag = 0;
        cor = data[i];
        if(i == 0){
            for(int j = 0; j<pointsDimension; j++){
                minarr[j] = cor[j];
                maxarr[j] = cor[j];
            }
        }else{
            for(int j = 0; j<pointsDimension; j++){
                if(cor[j] < minarr[j]){
                    minarr[j] = cor[j];
                }
                if(cor[j] > maxarr[j]){
                    maxarr[j] = cor[j];
                }
            }
        }
     }
}

float** run(float**data,int lnPoints,int lpointsDimension, float lt,int*nSample, float stdv){
  	
	// Parse part of the command-line parameters.
	nPoints = lnPoints;
    pointsDimension = lpointsDimension;

    float eps = lt;

	//default failure probability
	float delta = 0.001;

	//to count the running time
	MTimer verybegin;
    verybegin.go();
	
	initRandom();
	int originalnPoints;

	float* minarr;
    float* maxarr;
	float** sample;
    FAILIF(NULL == (minarr = (float*)MALLOC(sizeof(float)*pointsDimension)));
    FAILIF(NULL == (maxarr = (float*)MALLOC(sizeof(float)*pointsDimension)));
	
	readDataSetFromArray(data,nPoints,minarr, maxarr);
    originalnPoints = nPoints;
	PPointT *originalDataSet = dataSetPoints;
	nPoints = originalnPoints;
	
	//sample size
	int k = ceil(4/eps*sqrt(log(4/eps))*log(4/eps));
	if(k >= nPoints){
 	   	printf("If you want to achieve eps, you don't need to do reduce, just use the original data\n");
		*nSample = nPoints;
	    return data;
	}
	
	//get the number of round we need to run the algorithm
	int nofr = log(nPoints/k)/log(2);
	dataSetPoints = originalDataSet;

	for(int v = 0; v<nofr;v++){

  		pmStruct = newPMatchStructure(nPoints, pointsDimension, stdv * eps * 100);

		//run log(1/eps)+2 times and each time match point in the grid
  		for(int i = 0; i<log(1/eps)+2; i++){
            constructHash(i,pmStruct,dataSetPoints,minarr,maxarr);
            getMatchPairs(pmStruct,dataSetPoints);
		}

		//match arbitrarily after log(1/t)+2 times
  		matchArb(pmStruct,dataSetPoints);  
		if(v != 0){
			free(dataSetPoints);
		}
		
		//get the one point from each pair as the data for the next round
		dataSetPoints =  getSamplePoint(pmStruct,originalDataSet,pointsDimension);
  		nPoints = pmStruct->nMPoints;
  		
		//in the last round just output the sample to file
		if(v == nofr-1){
			verybegin.stop();
        	verybegin.update();
			printf("running time = %f\n",verybegin.elapsed);
			sample = getFloatArrSample(dataSetPoints,nPoints,pointsDimension);
            *nSample = nPoints;
   		}
  		freePMatchStructure(pmStruct);
	}
 	freeDataPoints(originalDataSet,originalnPoints);
	free(minarr);
    free(maxarr);
 	return sample;
}



JNIEXPORT jobjectArray JNICALL Java_org_myorg_HadoopKernel_start (JNIEnv * env, jobject obj, jobjectArray dataarr, jfloat t, jfloat stdv) {
    	int nPoints = env -> GetArrayLength(dataarr);
        jfloatArray jdata =  (jfloatArray)env->GetObjectArrayElement(dataarr, 0);
        int dimension = env -> GetArrayLength(jdata);

        float **localArray;
        localArray = new float*[nPoints];
        for(int i=0; i<nPoints; ++i){
            jfloatArray oneDim = (jfloatArray)env->GetObjectArrayElement(dataarr, i);
            jfloat *jdata=env->GetFloatArrayElements(oneDim, 0);
            //allocate localArray[i] using len2
            localArray[i] = new float[dimension];
            for(int j=0; j<dimension; ++j) {
                localArray[i][j]= jdata[j];
            }
            env->ReleaseFloatArrayElements(oneDim,jdata, 0);
        }
        float ft = t;
        int numSample;
        printf("before run nPoints = %d, dimension = %d , epsilon = %f, stdv = %f\n",nPoints,dimension,ft,stdv);
        float **sample = run(localArray,nPoints,dimension,ft,&numSample,stdv);
        printf("after run number of sample = %d\n", numSample);
        // Get the float array class
        jclass floatArrayClass = env->FindClass("[F");

        // Check if we properly got the float array class
        if (floatArrayClass == NULL)
        {
            // Ooops
            return NULL;
        }
        // Create the returnable 2D array
        jobjectArray myReturnable2DArray = env->NewObjectArray((jsize) numSample, floatArrayClass, NULL);

        // Go through the first dimension and add the second dimension arrays
        for (unsigned int i = 0; i < numSample; i++)
        {
            jfloatArray floatArray = env->NewFloatArray(dimension);
            env->SetFloatArrayRegion(floatArray, (jsize) 0, (jsize) dimension, (jfloat*) sample[i]);
            env->SetObjectArrayElement( myReturnable2DArray, (jsize) i, floatArray);
            env->DeleteLocalRef(floatArray);
        }
        // Return a Java consumable 2D float array
		if(numSample < nPoints){
			//if we do need to do sample
        	for(int i = 0; i <numSample; i++){
            	free(sample[i]);
        	}
        	free(sample);
			printf("free sample\n");
		}
		//localArray is free during freeDataPoints(originalDataSet,originalnPoints)
        return myReturnable2DArray;
}
