#include <stdio.h>
#include <stdlib.h>
#include <sys/times.h>
#include "headers.h"


// The data set containing all the points.
PPointT *dataSetPoints = NULL;
// Number of points in the data set.
IntT nPoints = 0;
// The dimension of the points.
IntT pointsDimension = 0;
// RealT
RealT t;

PPointMatchStructT pmStruct;
/*
  Prints the usage of the LSHMain.
 */
void usage(char *programName){
  printf("Usage: %s #pts_in_data_set dimension  data_set_file \n", programName);
}
inline RealT diffclock(clock_t clock1,clock_t clock2)
{
    RealT diffticks=clock1-clock2;
    RealT diffms=(diffticks)/CLOCKS_PER_SEC;
    return diffms;
}

inline PPointT readPoint(FILE *fileHandle){
  PPointT p;
  RealT sqrLength = 0;
  FAILIF(NULL == (p = (PPointT)MALLOC(sizeof(PointT))));
  FAILIF(NULL == (p->coordinates = (RealT*)MALLOC(pointsDimension * sizeof(RealT))));
  for(IntT d = 0; d < pointsDimension; d++){
    FSCANF_REAL(fileHandle, &(p->coordinates[d]));
  }
  //fscanf(fileHandle, "%[^\n]", sBuffer);
  p->index = -1;
  return p;
}

// Reads in the data set points from <filename> in the array
// <dataSetPoints>. Each point get a unique number in the field
// <index> to be easily indentifiable.
void readDataSetFromFile(char *filename){
  FILE *f = fopen(filename, "rt");
  FAILIF(f == NULL);

  //fscanf(f, "%d %d ", &nPoints, &pointsDimension);
  //FSCANF_DOUBLE(f, &thresholdR);
  //FSCANF_DOUBLE(f, &successProbability);
  //fscanf(f, "\n");
  FAILIF(NULL == (dataSetPoints = (PPointT*)MALLOC(nPoints * sizeof(PPointT))));
  for(IntT i = 0; i < nPoints; i++){
    dataSetPoints[i] = readPoint(f);
    dataSetPoints[i]->index = i;
    //printf("index = %d, first cord = %f\n",i,dataSetPoints[i]->coordinates[0]);
  }
}

void freeDataPoints(PPointT* dataSetPoints, IntT nPoints){
	PPointT p;
	for(IntT i = 0; i<nPoints; i++){
		p = dataSetPoints[i];
		if(p){
			free(p->coordinates);
			free(p);
		}
	}
	free(dataSetPoints);
}

/*
void getDummyData(PPointT dataPoint, IntT pointsDimension){

	if(dummyChain != NULL){
		dataPoint = dummyChain->data;
		dummyChain = dummyChain->next;
	}else{
		FAILIF(NULL == (dataPoint = (PPointT)MALLOC(sizeof(PointT))));
		FAILIF(NULL == (dataPoint->coordinates = (RealT*)MALLOC(pointsDimension * sizeof(RealT))));
		for(IntT d = 0; d < pointsDimension; d++){
    		dataPoint->coordinates[d] = 10;
   		 }
	}
}
*/
int main(int nargs, char **args){
  
  clock_t begin;
  clock_t end;

  if(nargs < 3){
    usage(args[0]);
    exit(1);
  }
  FILE *fp;
  FILE *tfp;

  initRandom();
  // Parse part of the command-line parameters.
  nPoints = atoi(args[1]);
  pointsDimension = atoi(args[2]);
 
  readDataSetFromFile(args[3]);
  if(nargs == 3){
	t = 1/nPoints;
  }else{
	t = atof(args[4]);
  }
  char filename[100] = {0};
  strcpy(filename,"err_");
  strcat(filename,args[4]);
  fp=fopen(filename, "a"); 
  
  char tfilename[100] = {0};
  strcpy(tfilename,"time_");
  strcat(tfilename,args[4]);
  tfp=fopen(tfilename, "a"); 

 begin=clock();
 
PPointT *fullPoints = NULL;
PPointT *fullOrgPoints = NULL;
RealT eps = t;
eps = eps/2;
IntT k = 4/eps*sqrt(log(4/eps));
IntT tempk = ceil(log(k)/log(2));
IntT k1 = pow(2,tempk);
IntT k2 = pow(2,tempk-1);
//make k to be 2^i
if(abs(k - k1) > abs(k - k2)){
	k = k2;
}else{
	k = k1;
}
IntT numofset = nPoints/k;
DPRINTF3("k = %d, num of sets = %d\n", k, numofset);
//divide point sets into n/k sets
IntT sofSet = k;

PPointT *originalDataSet = dataSetPoints;
IntT originalnPoints = nPoints;
PPointT *reducedDataSet;

//make numofset to be 2^i;
IntT temp = ceil(log(originalnPoints)/log(2));
nPoints = pow(2,temp);
numofset = nPoints/k;
IntT diff = nPoints - originalnPoints;

DPRINTF4("nPoints = %d, numofset = %d, diff = %d\n",nPoints, numofset, diff);
IntT dummyPointValue= (IntT)(dataSetPoints[0]->coordinates[0]*100);
DPRINTF("dummy = %d\n",dummyPointValue);
//make a new dataset with dummy points
FAILIF(NULL == (fullPoints = (PPointT*)MALLOC(nPoints  * sizeof(PPointT))));
FAILIF(NULL == (fullOrgPoints = (PPointT*)MALLOC(nPoints  * sizeof(PPointT))));
for(IntT i = 0; i<originalnPoints; i++){
	fullPoints[i] = dataSetPoints[i];
	fullOrgPoints[i] = dataSetPoints[i];
}
for(IntT i = 0; i < diff; i++){
	FAILIF(NULL == (fullPoints[originalnPoints+i] = (PPointT)MALLOC(sizeof(PointT))));
	FAILIF(NULL == (fullPoints[originalnPoints+i]->coordinates = (RealT*)MALLOC(pointsDimension * sizeof(RealT))));
	for(IntT d = 0; d < pointsDimension; d++){
		fullPoints[originalnPoints+i]->coordinates[d] = dummyPointValue;
	}

	FAILIF(NULL == (fullOrgPoints[originalnPoints+i] = (PPointT)MALLOC(sizeof(PointT))));
    FAILIF(NULL == (fullOrgPoints[originalnPoints+i]->coordinates = (RealT*)MALLOC(pointsDimension * sizeof(RealT))));
    for(IntT d = 0; d < pointsDimension; d++){
        fullOrgPoints[originalnPoints+i]->coordinates[d] = dummyPointValue;
    }

	fullPoints[originalnPoints+i]->index = originalnPoints+i;
	fullOrgPoints[originalnPoints+i]->index = originalnPoints+i;
} 

//IntT numDummy = 0;
nPoints = sofSet*2;
while(numofset > 1){
	nPoints = sofSet*2;
	FAILIF(NULL == (dataSetPoints = (PPointT*)MALLOC(nPoints  * sizeof(PPointT))));
	for(IntT j = 0; j<8; j++){
  		
		//pair all pairs
		for(IntT w = 0; w<numofset; w+=2){
			//get the pairs
			IntT m,x;
			//DPRINTF4("j = %d , w = %d, numofset = %d\n",j,w,numofset);
			for(m = w*sofSet,x = 0; m<(w+2)*sofSet; m++,x++){
				dataSetPoints[x] = fullPoints[m];
			}

			//reduce each pairs
			pmStruct = newPMatchStructure(nPoints, pointsDimension, t);
			for(IntT i = 0; i<log(nPoints)+2; i++){
				//DPRINTF("Round = %d \n",i);
				constructHash(i,pmStruct,dataSetPoints);
				getMatchPairs(pmStruct,dataSetPoints);
			}
			matchArb(pmStruct,dataSetPoints);
			//DPRINTF4("j = %d, w = %d ,pmStruct->nMPoints = %d\n",j,w,pmStruct->nMPoints);
			reducedDataSet =  getSamplePoint(pmStruct,fullOrgPoints,pointsDimension,j,w);
			//put the reduced set int the fullPoints for next round
			//since the num of points are reduced, we just put them into the origial fullPoints
		
			//DPRINTF("nMPoint = %d\n", pmStruct->nMPoints);	
			for(IntT m = 0; m<pmStruct->nMPoints; m++){
				fullPoints[w/2*sofSet+m] = reducedDataSet[m];
				//DPRINTF3("index = %d,value = %f\n",w/2*sofSet+m,reducedDataSet[m]->coordinates[0]);
			}
			freePMatchStructure(pmStruct);
			free(reducedDataSet);
		}
		numofset = numofset/2;
		//in the following round, just use the point after the bound, since we could add the point to the end arbitually 
		if(numofset < 2){
			break;
		}
		//the next round is based on the reduced points
	}
	//after 8 times reduction, we will skip one round
	if(numofset < 2){
        break;
    }
	numofset = numofset/2;
	sofSet = sofSet*2;
	free(dataSetPoints);
}

//run the final round for the reduce to k, it could be multiple round,just like naive
FAILIF(NULL == (dataSetPoints = (PPointT*)MALLOC(sofSet  * sizeof(PPointT))));
nPoints = sofSet;
if(nPoints<=k){
	end =clock();
	nPoints = 0;
	for(IntT m = 0; m<sofSet; m++){
    	if(fullPoints[m]->coordinates[0] != dummyPointValue){
        	dataSetPoints[m] = fullPoints[nPoints];
        	nPoints++;
    	}
	}
	RealT* err = errEst(originalDataSet,originalnPoints,dataSetPoints,nPoints,pointsDimension,1000,-10,10);
	RealT stderror = eps*2;
    DPRINTF6("n = %d, k = %d, l_infinity = %f,  epsilon = %f, fact = %f\n", originalnPoints, k, err[1], stderror, err[1]/stderror);
    fprintf(fp, "%f ", err[1]);
    fprintf(fp,"\n");
    fprintf(tfp,"%f ",RealT(diffclock(end,begin)));
    fprintf(tfp,"\n");
    DPRINTF("time spend = %f\n", RealT(diffclock(end,begin)));

}
else{
	for(IntT m = 0; m<sofSet; m++){
        dataSetPoints[m] = fullPoints[m];
	}
	IntT nofr = log(sofSet/k)/log(2);
	DPRINTF("number of round = %d\n",nofr);

	for(IntT v = 0; v<nofr;v++){
		pmStruct = newPMatchStructure(nPoints, pointsDimension, t);
		for(IntT i = 0; i<log(nPoints)+2; i++){
			//DPRINTF("Round = %d \n",i);
			constructHash(i,pmStruct,dataSetPoints);
			getMatchPairs(pmStruct,dataSetPoints);
		}
		matchArb(pmStruct,dataSetPoints);

		if(v != nofr-1){
			free(dataSetPoints);
			dataSetPoints  =  getSamplePoint(pmStruct,fullOrgPoints,pointsDimension,9,9);
			nPoints = pmStruct->nMPoints;
		}
 		else{
			//compute the error
 			reducedDataSet =  getSamplePoint2(pmStruct,fullOrgPoints,pointsDimension,9,9,dummyPointValue);
			end=clock();
			RealT* err = errEst(originalDataSet,originalnPoints,reducedDataSet,pmStruct->nMPoints,pointsDimension,1000,-10,10);

 			RealT stderror = eps*2;
 			DPRINTF6("n = %d, k = %d, l_infinity = %f,  epsilon = %f, fact = %f\n", originalnPoints, k, err[1], stderror, err[1]/stderror); 
			fprintf(fp, "%f ", err[1]);
 			fprintf(fp,"\n");
			fprintf(tfp,"%f ",RealT(diffclock(end,begin)));
			fprintf(tfp,"\n");
			DPRINTF("time spend = %f\n", RealT(diffclock(end,begin)));
 		}
		freePMatchStructure(pmStruct);
	}
}
DPRINTF1("Done");
free(reducedDataSet);
freeDataPoints(originalDataSet,originalnPoints);
for(IntT i = 0; i < diff; i++){
		free(fullOrgPoints[originalnPoints+i]->coordinates); 
		free(fullOrgPoints[originalnPoints+i]);
}
free(fullOrgPoints);
free(fullPoints);
fclose(fp);
fclose(tfp);
return 0;
}
