package org.myorg;

import java.io.*;
import java.util.*;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.conf.*;
import org.apache.hadoop.io.*;
import org.apache.hadoop.mapred.*;
import org.apache.hadoop.util.*;
import org.apache.hadoop.mapreduce.Mapper.Context;



// updated 19 OCT
public class HadoopRandSamp extends Configured implements Tool {


    public static class Map extends MapReduceBase implements Mapper<LongWritable, Text, NullWritable, DoubleWritable> {

      double eps;
      double delta;
      double prob;
      int tPts;
      int nPts;
      int numMappers;
      Random r;
      Vector<Text> values;
      

      public void configure(JobConf job) {
        eps = (double) job.getFloat( "HadoopRandSamp.eps", -1 );
        delta = (double) job.getFloat( "HadoopRandSamp.delta", -1 );
        tPts = job.getInt( "HadoopRandSamp.tPts", -1 );
        numMappers = job.getNumMapTasks();
        values = new Vector<Text>();
        double size = 1.0d / ( eps * eps );
        size *= Math.log( 1.0d / delta );
        nPts = (int) Math.ceil( size );
 
        prob = (double) nPts / tPts;
        r = new Random();

        System.out.println( "nPts: " + nPts );
        System.out.println( "numMappers: " + numMappers );

      }

      public void map(LongWritable key, Text value, OutputCollector<NullWritable, DoubleWritable> output, Reporter reporter) throws IOException {
         if( eps < 0 || delta < 0 || tPts < 0 ) throw new IOException( "param not set correctly" );

         if( r.nextDouble() < prob ) {

            String parts[] = value.toString().split(" ");
            double fval = Double.parseDouble( parts[0] );

            output.collect( NullWritable.get(), new DoubleWritable( fval ) );
         }
      }

      public void close() {
      }

    }

    public static class Reduce extends MapReduceBase implements Reducer<NullWritable, DoubleWritable, NullWritable, DoubleWritable> {

      public void configure(JobConf job) {
      }

       public void reduce(NullWritable key, Iterator<DoubleWritable> values, OutputCollector<NullWritable, DoubleWritable> output, Reporter reporter) throws IOException {

          while (values.hasNext()) {
            output.collect( NullWritable.get(), new DoubleWritable( values.next().get() ) );
          }

       }

    }

    public int run(String[] args) throws Exception {
     
      if( args.length < 4 ) {
         System.out.println( "Usage: <infile> <outdir> -delta <delta> -eps <epsilon> -tpts <totalpointsindataset>" );
         return 1;
      }

      JobConf conf = new JobConf(getConf(), HadoopRandSamp.class);
      conf.setJobName("HadoopRandSamp");

      conf.setOutputKeyClass(NullWritable.class);
      conf.setOutputValueClass(DoubleWritable.class);
     
      conf.setMapperClass(Map.class);
      conf.setReducerClass(Reduce.class);
      conf.setNumReduceTasks(1);

      conf.setInputFormat(TextInputFormat.class);
      conf.setOutputFormat(TextOutputFormat.class);

      List<String> other_args = new ArrayList<String>();
      for (int i=0; i < args.length; ++i) {
        if( "-delta".equals(args[i]) ) {
          conf.setFloat( "HadoopRandSamp.delta", Float.parseFloat( args[++i] ) );
          System.out.println( "delta: " + args[i] );
        }
        if( "-eps".equals(args[i]) ) {
          conf.setFloat( "HadoopRandSamp.eps", Float.parseFloat( args[++i] ) );
          System.out.println( "eps: " + args[i] );
        }
        if( "-tpts".equals(args[i]) ) {
          conf.setInt( "HadoopRandSamp.tPts", Integer.parseInt( args[++i] ) );
          System.out.println( "tpts: " + args[i] );
        }
        else {
          other_args.add(args[i]);
        }
      }

      FileInputFormat.setInputPaths(conf, new Path(other_args.get(0)));
      FileOutputFormat.setOutputPath(conf, new Path(other_args.get(1)));
      RunningJob rj = JobClient.runJob(conf);

      return 0;
    }

    public static void main(String[] args) throws Exception {
      int res = ToolRunner.run(new Configuration(), new HadoopRandSamp(), args);
      System.exit(res);
    }
}
