package org.myorg;

import java.io.*;
import java.util.*;

import org.apache.hadoop.fs.Path;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.conf.*;
import org.apache.hadoop.io.*;
import org.apache.hadoop.mapred.*;
import org.apache.hadoop.util.*;
import org.apache.hadoop.mapreduce.Mapper.Context;


public class HadoopSortSelect extends Configured implements Tool {

    public native void constructSummary( double eps, double delta, double[] data );   // for mappers
    public native double[] getBuf();                                              
    public native double[][] getLayer();
    public native void constructSummary( double eps, double delta, double[] inbuf, double[][] inlay );                    // for reducer
    public native void merge( double eps, double delta, double[] inbuf, double[][] inlay );                               // for reducer
    public native double valueAtRank( long rank );                                                                     // for reducer

    public static class Map extends MapReduceBase implements Mapper<LongWritable, Text, ByteWritable, SummaryContainer> {

      int mapId;
      double eps;
      double delta;
      String libFName;
      OutputCollector<ByteWritable, SummaryContainer> oc;

      Vector<Double> points;      

      public void configure(JobConf job) {

        mapId = job.getInt( "mapred.task.partition", -999 );
        eps = job.getFloat( "zKernel.eps", -1 );
        delta = job.getFloat( "zKernel.delta", -1 );

        System.out.println( "mapId: " + mapId );
        System.out.println( "eps: " + eps );
        System.out.println( "delta: " + delta );

        points = new Vector<Double>();

        Path[] libFiles = new Path[0];
        try {
          libFiles = DistributedCache.getLocalCacheFiles(job);
        } catch (IOException ioe) {
          System.err.println("Caught exception while getting cached files: " + StringUtils.stringifyException(ioe));
        }

        for( Path pathF : libFiles ) {
          libFName = pathF.toString();
          System.out.println( libFName );
        }
      }


      public void map(LongWritable key, Text value, OutputCollector<ByteWritable, SummaryContainer> output, Reporter reporter) throws IOException {
         if( eps < 0 || delta < 0 ) throw new IOException( "param not set correctly" );

         oc = output;

         String parts[] = value.toString().split(" ");
         Double fval = new Double( Double.parseDouble( parts[0] ) );
         points.add( fval );
      }

      public void close() {

        try {

           System.load(libFName);
           HadoopSortSelect myLib = new HadoopSortSelect();
           double[] buf = null;
           double[][] lay = null;

           double[] tdata = new double[points.size()];
           for( int i = 0; i < tdata.length; ++i )
              tdata[i] = points.get(i).doubleValue();

           myLib.constructSummary( eps, delta, tdata );
           buf = myLib.getBuf();
           lay = myLib.getLayer();

           if( buf == null ) System.out.println( "buf is null" );
           if( lay == null ) System.out.println( "lay is null" );

           DoubleWritable[] bufw = new DoubleWritable[buf.length];
           for( int i = 0; i < buf.length; ++i ) { 
              bufw[i] = new DoubleWritable( buf[i] );
           }

           DoubleWritable[][] layw = new DoubleWritable[lay.length][];
           for( int i = 0; i < layw.length; ++i ) {
              layw[i] = new DoubleWritable[lay[i].length];
              for( int j = 0; j < layw[i].length; ++j ) {
                 layw[i][j] = new DoubleWritable( lay[i][j] );
              }
           }
           DoubleArrayWritable[] faw = new DoubleArrayWritable[layw.length];
           for( int i = 0; i < layw.length; ++ i )
              faw[i] = new DoubleArrayWritable( layw[i] );
           
           oc.collect( new ByteWritable((byte) 1), new SummaryContainer( new DoubleArrayWritable( bufw ), faw ) );

        } catch(Exception e) {System.out.println( e.toString());}
      }
    }



    public static class Reduce extends MapReduceBase implements Reducer<ByteWritable, SummaryContainer, NullWritable, DoubleWritable> {

       double eps = 0;
       double delta = 0;
       int count = 0;
       int tpts = 0;
       String libFName;
       OutputCollector<NullWritable,DoubleWritable> oc;

       HadoopSortSelect myLib = null;
 
       public void configure( JobConf job ) {

          delta = job.getFloat( "zKernel.delta", 0 );
          eps = job.getFloat( "zKernel.eps", 0 );
          tpts = job.getInt( "zKernel.tpts", 0 );
          myLib = null;

          System.out.println( "delta: " + delta );
          System.out.println( "eps: " + eps );

          Path[] libFiles = new Path[0];
          try {
            libFiles = DistributedCache.getLocalCacheFiles(job);
          } catch (IOException ioe) {
            System.err.println("Caught exception while getting cached files: " + StringUtils.stringifyException(ioe));
          }
  
          for( Path pathF : libFiles ) {
            libFName = pathF.toString();
            System.out.println( libFName );
          }
        }

 
       public void reduce(ByteWritable key, Iterator<SummaryContainer> values, OutputCollector<NullWritable, DoubleWritable> output, Reporter reporter) throws IOException {
          if( eps < 0 || delta < 0 || tpts < 0 ) throw new IOException( "param not set correctly" );

          oc = output;
       
          while (values.hasNext()) {
             ++count;
             SummaryContainer sc = values.next();

             DoubleWritable[] fwbuff = (DoubleWritable[]) sc.buffer.toArray();
             double[] fbuff = new double[fwbuff.length];
             for( int i = 0; i < fbuff.length; ++i )
                fbuff[i] = fwbuff[i].get();
 
             DoubleWritable[][] fwlayer = new DoubleWritable[sc.layers.length][];
             for( int i = 0; i < sc.layers.length; ++i )
                fwlayer[i] = (DoubleWritable[]) sc.layers[i].toArray();
            
             double[][] flayers = new double[fwlayer.length][];
             for( int i = 0; i < flayers.length; ++i ) {
                flayers[i] = new double[fwlayer[i].length];
                for( int j = 0; j < flayers[i].length; ++j ) {
                   flayers[i][j] = fwlayer[i][j].get();
                }
             }

             if( myLib == null ) {
                System.load(libFName);
                myLib = new HadoopSortSelect();
                myLib.constructSummary( eps, delta, fbuff, flayers );
                System.out.println( "new summary");
             }
             else {
                myLib.merge( eps, delta, fbuff, flayers );
                System.out.println( "merged" );
             }

          }

      }


      public void close() {
         myLib.valueAtRank(-1);    

         int T = (int) Math.ceil( 1.0 / (double) eps );
         for( int j = 1; j <= T; ++j ) {
            double fj = (double) j;
            int index = (int) Math.ceil( (fj - 0.5) * ( (double) tpts / (double) T ) );
            try {
               oc.collect( NullWritable.get(), new DoubleWritable( myLib.valueAtRank(index) ) );
            }
            catch( Exception e ){}
         } 
      }
    }


    public int run(String[] args) throws Exception {

      if( args.length < 2 ) {
         System.out.println( "Usage: <infile> <outdir> -lib <libFile> -eps <epsilon> -delta <delta> -tpts <totalpts>" );
         return 1;
      }

      JobConf conf = new JobConf(getConf(), HadoopSortSelect.class);
      conf.setJobName("HadoopSortSelect");
      conf.set( "mapred.child.java.opts", "-server -Xmx4096m -Djava.net.preferIPv4Stack=true" );
      conf.setInt( "mapred.tasktracker.map.tasks.maximum", 1 );
      conf.setInt( "mapred.tasktracker.reduce.tasks.maximum", 1 );

      conf.setMapOutputKeyClass(ByteWritable.class);
      conf.setMapOutputValueClass(SummaryContainer.class);
      conf.setOutputKeyClass(NullWritable.class);
      conf.setOutputValueClass(DoubleWritable.class);
     
      conf.setMapperClass(Map.class);
      conf.setReducerClass(Reduce.class);
      conf.setNumReduceTasks(1);

      conf.setInputFormat(TextInputFormat.class);
      conf.setOutputFormat(TextOutputFormat.class);

      List<String> other_args = new ArrayList<String>();
      for (int i=0; i < args.length; ++i) {
        if ("-lib".equals(args[i])) {
          DistributedCache.addCacheFile(new Path(args[++i]).toUri(), conf);
          System.out.println( "lib: " + args[i] );
        }
        else if ("-eps".equals(args[i])) {
          conf.setFloat( "zKernel.eps", Float.parseFloat( args[++i] ) );
          System.out.println( "eps: " + args[i] );
        }
        else if( "-delta".equals(args[i]) ) {
          conf.setFloat( "zKernel.delta", Float.parseFloat( args[++i] ) );
          System.out.println( "delta: " + args[i] );
        }
        else if( "-tpts".equals(args[i]) ) {
          conf.setInt( "zKernel.tpts", Integer.parseInt( args[++i] ) );
          System.out.println( "tpts: " + args[i] );
        }      
        else {
          other_args.add(args[i]);
        }
      }

      FileInputFormat.setInputPaths(conf, new Path(other_args.get(0)));
      FileOutputFormat.setOutputPath(conf, new Path(other_args.get(1)));
      RunningJob rj = JobClient.runJob(conf);

      return 0;
    }

    public static void main(String[] args) throws Exception {
      int res = ToolRunner.run(new Configuration(), new HadoopSortSelect(), args);
      System.exit(res);
    }
}
