#include <iostream>
#include <fstream>
#include <cmath>
#include <algorithm>
#include <functional>
#include <cassert>
#include <stack>
#include <map>
#include <limits>
#include <sys/time.h>
#include <vector>
#include <cstring>
#include <cstdlib>
#include <ctime>
#include <climits>
#include "org_myorg_HadoopSortSelect.h"

using namespace std;



class QuantileEntry{
public:
    double v;
    long long rank;
    QuantileEntry() { }
    QuantileEntry(double _v, long long _rank)
        :v(_v), rank(_rank)
    { }
};

bool QuantileEntry_value_cmp (const QuantileEntry & qe1, const QuantileEntry & qe2)
{
    return qe1.v < qe2.v;
}

bool QuantileEntry_rank_cmp (const QuantileEntry & qe1, const QuantileEntry & qe2)
{
    return qe1.rank < qe2.rank;
}

class MergableQuantile
{
public:
    long long size_limit;                         // derived from eps and delta - does not need to be communicated
    std::vector<double> buffer, tmp;               // tmp only temporary swapping use
    std::vector<std::vector<double>> layers;       


    void init(double EPS, double DELTA)
    {
 
        if( DELTA >= EPS )
           size_limit =  (std::ceil(1.0 / EPS * std::sqrt(std::log(1.0 / EPS / DELTA) / std::log(2.0)) * (std::log(1.0 / EPS / DELTA) / std::log(2.0)) ));
        else
           size_limit =  (std::ceil(1.0 / EPS * std::sqrt(std::log(1.0 / EPS / DELTA) / std::log(2.0)) * (std::log(1.0 / DELTA) / std::log(2.0)) ));

        buffer.reserve(size_limit);
        tmp.reserve(size_limit);
    }



    void generate_from_data(const std::vector<double> & sorted_data_list)
    {
        buffer.clear();
        layers.clear();

        for(auto iter = sorted_data_list.begin(); iter != sorted_data_list.end(); ++iter)
        {
            feed(*iter);
        }
    }

    void feed(double v) {
        buffer.push_back(v);
        if(buffer.size() == size_limit)
        {
            std::sort(buffer.begin(), buffer.end());
            merge_in(buffer, 0);
            buffer.clear();
        }
    }

    void merge_one(std::vector<double> & dest, const std::vector<double> & v1, const std::vector<double> & v2)
    {
        auto iter1 = v1.begin();
        auto iter2 = v2.begin();
        bool use = (rand() % 2 == 0);
        dest.clear();
        while(true)
        {
            if((*iter1) < (*iter2))
            {
                if(use = !use) dest.push_back(*iter1);
                if(++iter1 == v1.end())
                {
                    while(iter2 != v2.end())
                    {
                        if(use = !use) dest.push_back(*iter2);
                        ++iter2;
                    }
                    break;
                }
            }
            else
            {
                if(use = !use) dest.push_back(*iter2);
                if(++iter2 == v2.end())
                {
                    while(iter1 != v1.end())
                    {
                        if(use = !use) dest.push_back(*iter1);
                        ++iter1;
                    }
                    break;
                }
            }
        }
    }

    void merge_in(std::vector<double> & new_v, int layer)
    {
        if(new_v.size() == 0) return;

        int idx = layer;
        while(true)
        {
            if(layers.size() <= idx)
            {
                layers.resize(idx+1);
                layers[idx].swap(new_v);
                break;
            }
            
            auto & v = layers[idx];
            if(v.size() == 0)
            {
                v.swap(new_v);
                break;
            }

            tmp.clear(); 
            tmp.swap(new_v);

            merge_one(new_v, v, tmp);

            v.clear();
            ++idx;
        }
    }

    void merge(MergableQuantile & mq)
    {
        for(int i = 0; i <(int) mq.layers.size(); ++i)
        {
            merge_in(mq.layers[i], i); 
        }

        if(buffer.size() + mq.buffer.size() < size_limit)
        {
            buffer.insert(buffer.end(), mq.buffer.begin(), mq.buffer.end());
        }
        else
        {
            auto iter = mq.buffer.begin() + (size_limit - buffer.size());
            buffer.insert(buffer.end(), mq.buffer.begin(), iter);
            std::sort(buffer.begin(), buffer.end());
            merge_in(buffer, 0);
            buffer.clear();
            buffer.insert(buffer.end(), iter, mq.buffer.end());
        }
    }

    size_t size(void) const
    {
        size_t s = 1 + buffer.size();
        for(auto iter = layers.begin(); iter != layers.end(); ++iter)
        {
            if(!(iter->empty()))
            {
                s += 1 + iter->size();
            }
        }
        return s;
    }

    void flatten(std::vector<QuantileEntry> & summary) {
        heap.clear();
        heap.reserve(layers.size() + 1);

        if(!buffer.empty())
        {
            std::sort(buffer.begin(), buffer.end());
            heap.push_back(HeapEntry(buffer.begin(), buffer.end(), 1));
        }

        {
            long long s = buffer.size();
            long long multiplier = 1;
            for(auto & v : layers)
            {
                if(!v.empty())
                {
                    heap.push_back(HeapEntry(v.begin(), v.end(), multiplier));
                    s += v.size();
                }
                multiplier *= 2;
            }
            summary.clear();
            summary.reserve(s);
        }

        auto comp = [](const HeapEntry & he1, const HeapEntry & he2) -> bool {
            return (*he2.iter) < (*he1.iter);
        };

        std::make_heap(heap.begin(), heap.end(), comp);

        long long cur_rank = 0;
        while(!heap.empty())
        {
            std::pop_heap(heap.begin(), heap.end(), comp);
            auto & he = heap.back();
            summary.push_back(QuantileEntry(*he.iter, cur_rank));
            cur_rank += he.multiplier;

            if(++he.iter == he.iterend)
            {
                heap.pop_back();
            }
            else
            {
                std::push_heap(heap.begin(), heap.end(), comp);
            }
        }
    }


    class HeapEntry{
    public:
        std::vector<double>::iterator iter, iterend;
        long long multiplier;
        HeapEntry(const std::vector<double>::iterator & _iter, const std::vector<double>::iterator & _iterend, long long _multiplier)
            :iter(_iter), iterend(_iterend), multiplier(_multiplier) { }
    };
    std::vector<HeapEntry> heap;
};

std::vector<QuantileEntry> g_summary;

double mergable_value_reporter(long rank)
{
    QuantileEntry pivot;
    pivot.rank = rank;
    auto iter = std::lower_bound(g_summary.begin(), g_summary.end(), pivot, QuantileEntry_rank_cmp);
    if(iter == g_summary.end())
        --iter;
    return iter->v; 
} 


int main()
{
//    MergableQuantile mq(0.1, 0.1 );
//    for(int i = 0; i < 10; ++i)
//    {
//        std::vector<double> v;
//        for(int j = 0; j < 100; j+=10)
//            v.push_back(j+i);
//
//        MergableQuantile tmp(0.1, 0.1);
//        tmp.generate_from_data(v);
//        mq.merge(tmp);
//    }
//
//    mq.flatten(g_summary);
//    for(int i = 0; i < 100; i += 10)
//        cout << i << ' ' << mergable_value_reporter(i) << endl;
//
//    return 0;
}


MergableQuantile mqGlob;

//    public native void constructSummary( double eps, double delta, double[] data,  double[] outbuf, double[][] outlay );
//    public native void constructSummary( double eps, double delta, double[] inbuf, double[][] inlay );
//    public native void merge( double[] inbuf, double[][] inlay );
//    public native double valueAtRank( long rank );


/*
 * Class:     org_myorg_HadoopSortSelect
 * Method:    constructSummary
 * Signature: (FF[F[F[[F)V
 */
JNIEXPORT void JNICALL Java_org_myorg_HadoopSortSelect_constructSummary__DD_3D
  (JNIEnv * env, jobject obj, jdouble eps, jdouble delta, jdoubleArray data) {

     mqGlob.init(eps, delta);
     std::vector<double> datav;

     jdouble* farr = env->GetDoubleArrayElements(data, 0);
     for( int i = 0; i < env->GetArrayLength(data); ++i )
        datav.push_back( farr[i] );
     env->ReleaseDoubleArrayElements(data,farr,0);

     mqGlob.generate_from_data(datav);
}


/*
 * Class:     org_myorg_HadoopSortSelect
 * Method:    getBuf
 * Signature: ()[F
 */
JNIEXPORT jdoubleArray JNICALL Java_org_myorg_HadoopSortSelect_getBuf
  (JNIEnv * env, jobject obj) {
     jdoubleArray outbuf = env->NewDoubleArray( mqGlob.buffer.size() );
     for( int i = 0; i < mqGlob.buffer.size(); ++i ) {
        env->SetDoubleArrayRegion( outbuf, (jsize) i, (jsize) 1, (jdouble*) &(mqGlob.buffer[i]) );
     }

     return outbuf;
}

/*
 * Class:     org_myorg_HadoopSortSelect
 * Method:    getLayer
 * Signature: ()[[F
 */
JNIEXPORT jobjectArray JNICALL Java_org_myorg_HadoopSortSelect_getLayer
  (JNIEnv *env, jobject obj) {
     jobjectArray outlay = env->NewObjectArray((jsize) mqGlob.layers.size(), env->FindClass("[D") , 0);
     for( int i = 0; i < mqGlob.layers.size(); ++i ) {
        jdoubleArray onelayer = env->NewDoubleArray((jsize) mqGlob.layers[i].size());
        for( int j = 0; j < mqGlob.layers[i].size(); ++j ) {
           env->SetDoubleArrayRegion(onelayer, (jsize) j, (jsize) 1, (jdouble*) &(mqGlob.layers[i][j]) );
        }
        env->SetObjectArrayElement( outlay, (jsize) i, onelayer );
        env->DeleteLocalRef(onelayer);
     } 
     return outlay;
}


/*
 * Class:     org_myorg_HadoopSortSelect
 * Method:    constructSummary
 * Signature: (FF[F[[F)V
 */
JNIEXPORT void JNICALL Java_org_myorg_HadoopSortSelect_constructSummary__DD_3D_3_3D
  (JNIEnv *env, jobject obj, jdouble eps, jdouble delta, jdoubleArray inbuf, jobjectArray inlay) {
   mqGlob.init(eps,delta);

   jdouble* farr = env->GetDoubleArrayElements(inbuf, 0);
   for( int i = 0; i < env->GetArrayLength( inbuf ); ++i )
      mqGlob.buffer.push_back( farr[i] );
   env->ReleaseDoubleArrayElements(inbuf,farr,0);


   mqGlob.layers.resize( env->GetArrayLength(inlay) );
   for( int i = 0; i < env->GetArrayLength(inlay); ++i ) {
      jdoubleArray onelayer = (jdoubleArray) env->GetObjectArrayElement(inlay,i);
      farr = env->GetDoubleArrayElements(onelayer,0);
      for( int j = 0; j < env->GetArrayLength(onelayer); ++j )
         mqGlob.layers[i].push_back( farr[j] );
 
      env->ReleaseDoubleArrayElements(onelayer, farr, 0 );
   }

}

/*
 * Class:     org_myorg_HadoopSortSelect
 * Method:    merge
 * Signature: ([F[[F)V
 */
JNIEXPORT void JNICALL Java_org_myorg_HadoopSortSelect_merge
  (JNIEnv *env, jobject obj, jdouble eps, jdouble delta, jdoubleArray inbuf, jobjectArray inlay) {

   MergableQuantile mq;
   mq.init(eps,delta);

   jdouble* farr = env->GetDoubleArrayElements(inbuf, 0);
   for( int i = 0; i < env->GetArrayLength( inbuf ); ++i )
      mq.buffer.push_back( farr[i] );
   env->ReleaseDoubleArrayElements(inbuf,farr,0);

   mq.layers.resize( env->GetArrayLength(inlay) );
   for( int i = 0; i < env->GetArrayLength(inlay); ++i ) {
      jdoubleArray onelayer = (jdoubleArray) env->GetObjectArrayElement(inlay,i);
      farr = env->GetDoubleArrayElements(onelayer,0);
      for( int j = 0; j < env->GetArrayLength(onelayer); ++j )
         mq.layers[i].push_back( farr[j] );
 
      env->ReleaseDoubleArrayElements(onelayer, farr, 0 );
   }

   mqGlob.merge(mq);

}

/*
 * Class:     org_myorg_HadoopSortSelect
 * Method:    valueAtRank
 * Signature: (J)F
 */
JNIEXPORT jdouble JNICALL Java_org_myorg_HadoopSortSelect_valueAtRank
  (JNIEnv * env, jobject obj, jlong rank){

   if( rank < 0 ) mqGlob.flatten(g_summary);
   else return mergable_value_reporter(rank);

}


