package sort;
import java.io.*;
import java.util.*;
import java.util.Date;
import java.util.Random;
import java.util.ArrayList;
import java.math.*;


public class Zorder {
	private static String createExtra(int num) {
		if( num < 1 ) return "";

		char[] extra = new char[num];
		for (int i = 0; i < num; i++) 
			extra[i] = '0';
		return (new String(extra)); 	
	}

	// Create random shifted vector for zorder computation	
	public static int[] createShift(int dimension, Random rin, boolean shift) {
		Random r = rin; 
		int[] rv = new int[dimension];  // random vector 

		if (shift) {
			for (int i = 0; i < dimension; i++) {
				rv[i] = ((int) Math.abs(r.nextInt())); // % range;
			}	
		} else {
			for (int i = 0; i < dimension; i++)	
				rv[i] = 0;
		}

		return rv;
	}

	// scale and convert to int first
	public static int[] shift( int[] coord, int[] shift ) {
		int[] result = new int[coord.length];
		for( int i = 0; i < coord.length; ++i )
			result[i] = coord[i] + shift[i];

		return result;
	}

        // Deprecated -> use valueOf(dimension,coord) instead
	// Convert an multi-dimensional coordinate into a zorder
	public static String valueOf(int dimension, float[] coord, int scale, int[] rs) {
		ArrayList<String> arrPtr = new ArrayList<String>(dimension);
		ArrayList<String> arrPtr1 = new ArrayList<String>(dimension);
		int max = 0;
		int fix = 15; //global maximum possible zvalue length

		for (int i = 0; i < dimension; i++) {
			String p = Integer.toBinaryString((int) (coord[i] * scale) 
					+ rs[i]);
			if (p.length() > max)
				max = p.length();
			arrPtr.add(p);
		}

		for (String e: arrPtr) {
			String extra = createExtra(max - e.length()); 
            arrPtr1.add(extra + e);
		}
	
		char[] value = new char[dimension * max];
		int index = 0;

		// Create Zorder
		for (int i = 0; i < max; i++) {
			for (String e: arrPtr1) {
				char ch = e.charAt(i);
				value[index++] = ch;
			}	
		}		
	
		String order = new String(value);
		long ret = Long.parseLong(order, 2);
		order = Long.toString(ret);	
		
		if (order.length() < fix) {
			String extra = createExtra(fix - order.length());
			order = extra + order;
		}
		
		return order;
	}


	public static int maxDecDigits( int dimension ) {
		int max = 32;
		BigInteger maxDec = new BigInteger( "1" );
		maxDec = maxDec.shiftLeft( dimension * max );
		maxDec.subtract( BigInteger.ONE );
		return maxDec.toString().length();
	}

	public static String maxDecString( int dimension ) {
		int max = 32;
		BigInteger maxDec = new BigInteger( "1" );
		maxDec = maxDec.shiftLeft( dimension * max );
		maxDec.subtract( BigInteger.ONE );
		return maxDec.toString();
	}
	

	// Convert an multi-dimensional coordinate into a zorder
	// coordinates have already been scaled and shifted
	public static String valueOfF(int dimension, float[] coord) {
		Vector<String> arrPtr = new Vector<String>(dimension);
        int max = 32;
		int fix = maxDecDigits(dimension); //global maximum possible zvalue length

		for (int i = 0; i < dimension; i++) {
			String p = Integer.toBinaryString((int)coord[i]);
			arrPtr.add(p);
		}

		for( int i = 0; i < arrPtr.size(); ++i ) {
			String extra = createExtra( max - arrPtr.elementAt(i).length() ); 
			arrPtr.set(i, extra + arrPtr.elementAt(i) );
		}
	
		char[] value = new char[dimension * max];
		int index = 0;

		// Create Zorder
		for (int i = 0; i < max; ++i ) {
			for (String e: arrPtr) {
				char ch = e.charAt(i);
				value[index++] = ch;
			}	
		}		
			
		String order = new String(value);
		// Covert a binary representation of order into a big integer
		BigInteger ret = new BigInteger( order, 2 );

		// Return a fixed length decimal String representation of 
		// the big integer (z-order)
		order = ret.toString();	
		if (order.length() < fix) {
			String extra = createExtra(fix - order.length());
			order = extra + order;
		} else if (order.length() > fix) {
			System.out.println("too big zorder, need to fix Zorder.java");
			System.exit(-1);
		}

		
		return order;
	}


	// Convert an multi-dimensional coordinate into a zorder
	// coordinates have already been scaled and shifted
	public static String valueOf(int dimension, int[] coord) {
		Vector<String> arrPtr = new Vector<String>(dimension);
        int max = 32;
		int fix = maxDecDigits(dimension); //global maximum possible zvalue length

		for (int i = 0; i < dimension; i++) {
			if( coord[i] < 0 ) System.out.println( "WARNING: negative value " + coord[i] + "!!!" );
			String p = Integer.toBinaryString((int)coord[i]);
			arrPtr.add(p);
		}

		for( int i = 0; i < arrPtr.size(); ++i ) {
			String extra = createExtra( max - arrPtr.elementAt(i).length() ); 
			arrPtr.set(i, extra + arrPtr.elementAt(i) );
		}
	
		char[] value = new char[dimension * max];
		int index = 0;

		// Create Zorder
		for (int i = 0; i < max; ++i ) {
			for (String e: arrPtr) {
				char ch = e.charAt(i);
				value[index++] = ch;
			}	
		}		
			
		String order = new String(value);
		// Covert a binary representation of order into a big integer
		BigInteger ret = new BigInteger( order, 2 );

		// Return a fixed length decimal String representation of 
		// the big integer (z-order)
		order = ret.toString();	
		if (order.length() < fix) {
			String extra = createExtra(fix - order.length());
			order = extra + order;
		} else if (order.length() > fix) {
			System.out.println("too big zorder, need to fix Zorder.java");
			System.exit(-1);
		}

		
		return order;
	}

	public static int[] toCoord(String z, int dimension) {
		int DECIMAL_RADIX = 10;
		int BINARY_RADIX = 2;

		if (z == null) {
			System.out.println("Z-order Null pointer!!!@Zorder.toCoord");
			System.exit(-1);	
		}

		BigInteger bigZ = new BigInteger(z, DECIMAL_RADIX);
		String bigZStr = bigZ.toString(BINARY_RADIX);


		int len = bigZStr.length();
		int prefixZeros = len % dimension;
		String prefix = Zorder.createExtra(prefixZeros);
		bigZStr = prefix + bigZStr;
		
		len = bigZStr.length();
		if (len % dimension != 0) {
			System.out.println("Wrong prefix!!!@Zorder.toCoord");
			System.exit(-1);
		}

		//The most significant bit is save at starting position of 
		//the char array.
		char[] bigZCharArray = bigZStr.toCharArray();

		int[] coord = new int[dimension];
		for (int i = 0; i < dimension; i++)
			coord[i] = 0;



                for( int i = 0; i < bigZCharArray.length; ) {
                   for( int j = 0; j < dimension; ++j ) {
                      coord[j] <<= 1;
                      coord[j] |= bigZCharArray[i++] - '0';
                   }
                }

		return coord;
	}

	public static float[] toCoordf(String z, int dimension) {
		int DECIMAL_RADIX = 10;
		int BINARY_RADIX = 2;

		if (z == null) {
			System.out.println("Z-order Null pointer!!!@Zorder.toCoord");
			System.exit(-1);	
		}

		BigInteger bigZ = new BigInteger(z, DECIMAL_RADIX);
		String bigZStr = bigZ.toString(BINARY_RADIX);


		int len = bigZStr.length();
		int prefixZeros = len % dimension;
		String prefix = Zorder.createExtra(prefixZeros);
		bigZStr = prefix + bigZStr;
		
		len = bigZStr.length();
		if (len % dimension != 0) {
			System.out.println("Wrong prefix!!!@Zorder.toCoord");
			System.exit(-1);
		}

		//The most significant bit is save at starting position of 
		//the char array.
		char[] bigZCharArray = bigZStr.toCharArray();

		int[] coord = new int[dimension];
		for (int i = 0; i < dimension; i++)
			coord[i] = 0;



                for( int i = 0; i < bigZCharArray.length; ) {
                   for( int j = 0; j < dimension; ++j ) {
                      coord[j] <<= 1;
                      coord[j] |= bigZCharArray[i++] - '0';
                   }
                }

		//make a float array and return it
		float[] float_coord = new float[dimension];
        for( int j = 0; j < dimension; ++j )
		   float_coord[j] = coord[j];	
			
		//return coord;
		return float_coord;
	}

	public static int toScaledInt( float f ) {		
		int scale = 1000;

		int intpart = (int) f;
		float floatpart = f;

		floatpart = floatpart - intpart;
		intpart *= scale;
		intpart += (int) ( floatpart * scale );

		return intpart;
	}

	public static void toZvalFile( int dimension, String inFile, String outFile ) {
		final int bufferSize = 16 * 1024 * 1024;
		String line = "";

		try {
			BufferedReader rin = new BufferedReader(new FileReader(inFile),bufferSize);
			BufferedWriter rout = new BufferedWriter(new FileWriter(outFile),bufferSize);

			while( ( line = rin.readLine() ) != null ) {
				String[] parts = line.split( " " );
				if( parts.length < dimension ) continue;

				int[] coord = new int[dimension];
				for( int i = 0; i < dimension; ++i ) {
					float f = Float.valueOf( parts[i] );
					coord[i] = Zorder.toScaledInt( f );
				}
				String ZVal = Zorder.valueOf(dimension, coord);
				rout.write(ZVal + " " + line);
				rout.newLine();

			}

			rin.close();
			rout.close();
		}
		catch(IOException e )
		{} 
	}

	public static void toBinaryZvalFile( int dimension, String inFile, String outFile, String shiftFile ) {
		final int bufferSize = 16 * 1024 * 1024;
		String line = "";

		int[] shiftV = new int[dimension];
		for( int i = 0; i < shiftV.length; ++i )
			shiftV[i] = 0;

		try {

			if( shiftFile != null ) {
				Scanner fs = new Scanner( new BufferedInputStream( new FileInputStream( shiftFile ), 8 * 1024 * 1024 ) );
				String sline = fs.nextLine();
				String[] parts = sline.split( " " );

				for( int i = 0; i < shiftV.length; ++i ) {
					shiftV[i] = Integer.parseInt(parts[i]);
					System.out.println( shiftV[ i ] );
				}
			}

			BufferedReader rin = new BufferedReader(new FileReader(inFile),bufferSize);
			DataOutputStream dos = new DataOutputStream( new BufferedOutputStream( new FileOutputStream( outFile), bufferSize) );

			while( ( line = rin.readLine() ) != null ) {
				String[] parts = line.split( " " );
				if( parts.length < dimension ) continue;

				int[] coord = new int[dimension];
				float[] ocoord = new float[dimension];
				for( int i = 0; i < dimension; ++i ) {
					float f = Float.valueOf( parts[i] );
					coord[i] = Zorder.toScaledInt( f );
					ocoord[i] = f;
				}
				coord = shift( coord, shiftV );
				String ZVal = Zorder.valueOf(dimension, coord);
				ZorderSortRecord rec = new ZorderSortRecord( ZVal, ocoord );
				rec.write(dos);

			}

			rin.close();
			dos.close();
		}
		catch(IOException e )
		{} 
	}

	public static void binaryToAsciiFile( int dimension, int ZstringLength, String inFile, String outFile ) {
		final int bufferSize = 16 * 1024 * 1024;
		ZorderSortRecord zRec = new ZorderSortRecord(ZstringLength,dimension);
		DataInputStream dis;
		BufferedWriter bw;

		try {
			dis = new DataInputStream( new BufferedInputStream( new FileInputStream( inFile ), bufferSize ) );
			bw = new BufferedWriter( new FileWriter( outFile ), bufferSize );

			try {
				while( true ) {
					zRec.read( dis );
					zRec.write( bw );
				}
			}
			catch( Exception e ) {}

			dis.close();
			bw.close();

		}
		catch( Exception e ) {}
	}


	public static void binaryToAsciiFile( int dimension, int ZstringLength, int k, String inFile, String outFile ) {
		final int bufferSize = 16 * 1024 * 1024;
		ZorderSortRecord zRec = new ZorderSortRecord(ZstringLength,dimension);
		RandomAccessFile dis;
		BufferedWriter bw;
		long pos = 0;

		try {
			dis = new RandomAccessFile( inFile, "r" );
			bw = new BufferedWriter( new FileWriter( outFile ), bufferSize );


			try {
				zRec.read( dis );
				zRec.write( bw );
				long numRecords = (dis.length() / zRec.size() );
				long skipLength = (long) Math.floor( ( numRecords - k ) / ( k - 1 ) );

				System.out.println( "numRecs: " + numRecords );
				System.out.println( "skipLength: " + skipLength );
				skipLength *= zRec.size();

				while( true ) {
					dis.seek( dis.getFilePointer() + skipLength );
					zRec.read( dis );
					zRec.write( bw );
				}
			}
			catch( Exception e ) {}

			dis.close();
			bw.close();

		}
		catch( Exception e ) {}
	}

	public static void main(String[] args) {

		if( args.length < 4 ) {
			System.out.println( "usage: <dimension> <infile> <outfile> <type> [k] [shiftfile]" );
			System.out.println( "<type> = (0) Ascii to Zorder binary" );
			System.out.println( "<type> = (1) binary zorder to ascii zorder" );
			System.out.println( "<type> = (2) sample binary zorder to ascii zorder" );
			return;
		}

		int type = Integer.valueOf( args[3] );
		int dimension = Integer.valueOf( args[0] );
		String shiftFile = args.length == 6 ? args[5] : null;

		if( type == 0 ) {
			Zorder.toBinaryZvalFile( dimension, args[1], args[2], shiftFile );
		}
		else if( type == 1 ) {
			Zorder.binaryToAsciiFile(dimension, 0, args[1], args[2]);
		}
		else if( type == 2 ) {
			int k = Integer.valueOf(args[4] );
			Zorder.binaryToAsciiFile(dimension,0,k,args[1],args[2]);
		}

	}
}
